@@ -5,20 +5,31 @@ import pandas as pd
 from collections import Counter
 from pathlib import Path
 
-from .md_utils import (get_md_relpaths_matching_subdirs, get_unique_md_links,
-                       get_unique_wikilinks,
-                       get_md_links,
-                       get_wikilinks,
-                       get_embedded_files,
-                       get_front_matter,
+# init
+from .md_utils import (get_md_relpaths_matching_subdirs)
+from .canvas_utils import (get_canvas_relpaths_matching_subdirs)
+# connect
+from .md_utils import (_get_md_front_matter_and_content,
+                       _get_html_from_md_content,
+                       _get_md_links_from_source_text,
+                       _get_unique_md_links_from_source_text,
+                       _get_unique_wikilinks_from_source_text,
+                       _get_all_wikilinks_from_source_text,
+                       _get_all_embedded_files_from_source_text,
                        get_tags,
-                       _get_source_text_from_md_file,
-                       _get_readable_text_from_md_file,
-                       _get_all_latex_from_md_file)
+                       get_source_text_from_html,
+                       _get_all_latex_from_html_content)
+# gather
+from .md_utils import (get_source_text_from_md_file,
+                       get_readable_text_from_md_file)
+# canvas:
+from .canvas_utils import (get_canvas_content,
+                           get_canvas_graph_detail)
 
 
 class Vault:
-    def __init__(self, dirpath, *, include_subdirs=None, include_root=True):
+    def __init__(self, dirpath: Path, *,
+                 include_subdirs: list[str] = None, include_root: bool = True):
         """A Vault object lets you dig into your Obsidian vault, by giving
         you a toolkit for analysing its contents.  Specify a dirpath to
         instantiate the class.  This class is intended to support multiple
@@ -30,6 +41,7 @@ class Vault:
         counted rather than notes.  For example, if a note has 2 backlinks
         then there will be 2 wikilinks (to another note) visible when you do
         analysis of the edges & nodes.
+        Detail on canvas files is also stored when calling this method.
 
         By calling GATHER you store the plaintext of your vault in the notes
         index attribute.  You can specify rules for how the text is
@@ -63,47 +75,62 @@ class Vault:
             get_backlinks
             get_backlink_counts
             get_wikilinks
+            get_wikilink_counts
             get_embedded_files
             get_math
             get_front_matter
             get_md_links
             get_tags
             get_source_text
+            get_readable_text
 
         Methods for analysis across multiple notes:
             get_note_metadata
 
-        Attributes:
+        Attributes - general:
             dirpath (arg)
+            is_connected
+            is_gathered
+        Attributes - md-related:
             file_index
+            graph
             backlinks_index
             wikilinks_index
+            unique_wikilinks_index
             embedded_files_index
             math_index
             md_links_index
+            unique_md_links_index
             tags_index
             nonexistent_notes
             isolated_notes
-            graph
             source_text_index
             readable_text_index
-            is_connected
-            is_gathered
+        Attributes - canvas-related:
+            canvas_file_index
+            canvas_content_index
+            canvas_graph_detail_index
         """
         self._dirpath = dirpath
         self._file_index = self._get_md_relpaths_by_name(
             include_subdirs=include_subdirs,
             include_root=include_root)
+        self._canvas_file_index = self._get_canvas_relpaths_by_name(
+            include_subdirs=include_subdirs,
+            include_root=include_root)
 
-        # graph setup
-        self._graph = None
         self._is_connected = False
         self._is_gathered = False
+
+        # via md content:
+        self._graph = None
         self._backlinks_index = {}
         self._wikilinks_index = {}
+        self._unique_wikilinks_index = {}
         self._embedded_files_index = {}
         self._math_index = {}
         self._md_links_index = {}
+        self._unique_md_links_index = {}
         self._tags_index = {}
         self._nonexistent_notes = []
         self._isolated_notes = []
@@ -111,59 +138,89 @@ class Vault:
         self._source_text_index = {}
         self._readable_text_index = {}
 
+        # via canvas content:
+        self._canvas_content_index = {}
+        self._canvas_graph_detail_index = {}
+
     @property
-    def dirpath(self):
+    def dirpath(self) -> Path:
         """pathlib Path"""
         return self._dirpath
 
     @property
-    def file_index(self):
-        """dict: one-to-one mapping of filename (k) to relative path (v)"""
+    def file_index(self) -> dict[str, Path]:
+        """dict: one-to-one mapping of md filename (k) to relative path (v)"""
         return self._file_index
 
+    @file_index.setter
+    def file_index(self, value) -> dict[str, Path]:
+        self._file_index = value
+
     @property
-    def graph(self):
+    def canvas_file_index(self) -> dict[str, Path]:
+        """dict: one-to-one mapping of canvas filename (k) to relative path
+        (v)"""
+        return self._canvas_file_index
+
+    @canvas_file_index.setter
+    def canvas_file_index(self, value) -> dict[str, Path]:
+        self._canvas_file_index = value
+
+    @property
+    def graph(self) -> nx.MultiDiGraph:
         """networkx Graph"""
         return self._graph
 
     @property
-    def backlinks_index(self):
+    def backlinks_index(self) -> dict[str, list[str]]:
         """dict of lists: note name (k) to lists (v).  v is [] if k
         has no backlinks."""
         return self._backlinks_index
 
     @property
-    def wikilinks_index(self):
+    def wikilinks_index(self) -> dict[str, list[str]]:
         """dict of lists: filename (k) to lists (v).  v is [] if k
         has no wikilinks."""
         return self._wikilinks_index
 
     @property
-    def embedded_files_index(self):
-        """dict: note name (k) to list of embedded file strimg (v).
+    def unique_wikilinks_index(self) -> dict[str, list[str]]:
+        """dict of lists: filename (k) to lists (v).  v is [] if k
+        has no wikilinks."""
+        return self._unique_wikilinks_index
+
+    @property
+    def embedded_files_index(self) -> dict[str, list[str]]:
+        """dict: note name (k) to list of embedded file string (v).
         v is [] if k has no embedded files."""
         return self._embedded_files_index
 
     @property
-    def math_index(self):
-        """dict: note name (k) to list of LaTeX math strimg (v).  v is [] if
+    def math_index(self) -> dict[str, list[str]]:
+        """dict: note name (k) to list of LaTeX math string (v).  v is [] if
         k has no LaTeX."""
         return self._math_index
 
     @property
-    def md_links_index(self):
+    def md_links_index(self) -> dict[str, list[str]]:
         """dict of lists: filename (k) to lists (v).  v is [] if k
         has no markdown links."""
         return self._md_links_index
 
     @property
-    def tags_index(self):
+    def unique_md_links_index(self) -> dict[str, list[str]]:
+        """dict of lists: filename (k) to lists (v).  v is [] if k
+        has no markdown links."""
+        return self._unique_md_links_index
+
+    @property
+    def tags_index(self) -> dict[str, list[str]]:
         """dict of lists: filename (k) to lists (v).  v is [] if k
         has no tags."""
         return self._tags_index
 
     @property
-    def nonexistent_notes(self):
+    def nonexistent_notes(self) -> list[str]:
         """list: notes without files, i.e. the notes have backlink(s) but
         their md files don't exist yet.
 
@@ -172,30 +229,30 @@ class Vault:
         return self._nonexistent_notes
 
     @property
-    def isolated_notes(self):
+    def isolated_notes(self) -> list[str]:
         """list: notes (with their own md files) that lack backlinks and
         lack wikilinks.  They are not connected to other notes in the
         Obsidian graph at all."""
         return self._isolated_notes
 
     @property
-    def front_matter_index(self):
+    def front_matter_index(self) -> dict[str, list[str]]:
         """dict: note name (k) to front matter (v).  v is {} if no front
         matter was extracted from note."""
         return self._front_matter_index
 
     @property
-    def is_connected(self):
+    def is_connected(self) -> bool:
         """Bool: has the connect function been called to set up graph?"""
         return self._is_connected
 
     @property
-    def is_gathered(self):
+    def is_gathered(self) -> bool:
         """Bool: has the gather function been called to gather text?"""
         return self._is_gathered
 
     @property
-    def source_text_index(self):
+    def source_text_index(self) -> dict[str, str]:
         """dict of strings: filename (k) to source text string (v).  v is ''
         if k has no text.
 
@@ -204,8 +261,12 @@ class Vault:
         to maintain the additional markdown features that Obsidian uses."""
         return self._source_text_index
 
+    @source_text_index.setter
+    def source_text_index(self, value) -> dict[str, str]:
+        self._source_text_index = value
+
     @property
-    def readable_text_index(self):
+    def readable_text_index(self) -> dict[str, str]:
         """dict of strings: filename (k) to 'readable' string (v).  v is ''
         if k has no text.
 
@@ -214,35 +275,129 @@ class Vault:
         needed, while still maintaining the meaning of text."""
         return self._readable_text_index
 
-    def connect(self):
+    @readable_text_index.setter
+    def readable_text_index(self, value) -> dict[str, str]:
+        self._readable_text_index = value
+
+    @property
+    def canvas_content_index(self) -> dict[str, str]:
+        """dict of dict: 'shortest path when possible' filepath with canvas
+        ext (k), to canvas content dict (v).  v is {} if k has no content."""
+        return self._canvas_content_index
+
+    @canvas_content_index.setter
+    def canvas_content_index(self, value) -> dict[str, str]:
+        self._canvas_content_index = value
+
+    @property
+    def canvas_graph_detail_index(self) -> \
+        dict[str, tuple[nx.MultiDiGraph,
+                        dict[str, tuple[int, int]],
+                        dict[tuple[str, str], str]]
+             ]:
+        """dict of tuple: 'shortest path when possible' filepath with canvas
+        ext (k), to canvas graph detail tuple (v)."""
+        return self._canvas_graph_detail_index
+
+    @canvas_graph_detail_index.setter
+    def canvas_graph_detail_index(self, value) -> \
+        dict[str, tuple[nx.MultiDiGraph,
+                        dict[str, tuple[int, int]],
+                        dict[tuple[str, str], str]]
+             ]:
+        self._canvas_graph_detail_index = value
+
+    def connect(self, *, show_nested_tags: bool = False):
         """connect your notes together by representing the vault as a
         Networkx graph object, G.
 
         With your vault instantiated, set up the graph through:
             vault.connect()
 
-        The graph G is written to the 'graph' attribute.
+        The graph G is written to the 'graph' attribute, to represent the
+        Obsidian graph of your notes.
+
+        Args:
+            show_nested_tags (Boolean): show nested tags in the output.
+                Defaults to False (which would mean only the highest level
+                of any nested tags are included in the output).
         """
+        # md content:
         if not self._is_connected:
-            # default graph to mirror Obsidian's link counts
-            wiki_link_map = self._get_wikilinks_index()
-            G = nx.MultiDiGraph(wiki_link_map)
+            # index dicts, where k is a note name in the vault:
+            md_links_ix = {}
+            md_links_unique_ix = {}
+            embedded_files_ix = {}
+            tags_ix = {}
+            math_ix = {}
+            front_matter_ix = {}
+            wikilinks_ix = {}
+            wikilinks_unique_ix = {}
+
+            # loop through md files:
+            for f, relpath in self._file_index.items():
+                # MAIN file read:
+                front_matter, content = _get_md_front_matter_and_content(
+                    self._dirpath / relpath)
+                html = _get_html_from_md_content(content)
+                src_txt = get_source_text_from_html(
+                    html, remove_code=True)
+
+                # info from core text:
+                md_links_ix[f] = _get_md_links_from_source_text(src_txt)
+                md_links_unique_ix[f] = _get_unique_md_links_from_source_text(src_txt)
+                embedded_files_ix[f] = _get_all_embedded_files_from_source_text(
+                    src_txt, remove_aliases=True)
+                wikilinks_ix[f] = _get_all_wikilinks_from_source_text(
+                    src_txt, remove_aliases=True)
+                wikilinks_unique_ix[f] = _get_unique_wikilinks_from_source_text(
+                    src_txt, remove_aliases=True)
+                # info from html:
+                math_ix[f] = _get_all_latex_from_html_content(html)
+                # split out front matter:
+                front_matter_ix[f] = front_matter
+
+                # MORE file reads needed for extra info:
+                tags_ix[f] = get_tags(self._dirpath / relpath,
+                                      show_nested=show_nested_tags)
+
+            self._md_links_index = md_links_ix
+            self._unique_md_links_index = md_links_unique_ix
+            self._embedded_files_index = embedded_files_ix
+            self._tags_index = tags_ix
+            self._math_index = math_ix
+            self._front_matter_index = front_matter_ix
+            # to be used for graph:
+            self._wikilinks_index = wikilinks_ix
+            self._unique_wikilinks_index = wikilinks_unique_ix
+
+            # graph:
+            G = nx.MultiDiGraph(wikilinks_ix)
             self._graph = G
+            # info obtained from graph:
             self._backlinks_index = self._get_backlinks_index(graph=G)
-            self._wikilinks_index = wiki_link_map
-            self._md_links_index = self._get_md_links_index()
-            self._tags_index = self._get_tags_index()
             self._nonexistent_notes = self._get_nonexistent_notes()
             self._isolated_notes = self._get_isolated_notes(graph=G)
-            self._embedded_files_index = self._get_embedded_files_index()
-            self._math_index = self._get_math_index()
-            self._front_matter_index = self._get_front_matter_index()
 
             self._is_connected = True
 
+        # canvas content:
+        # loop through canvas files:
+        canvas_content_ix = {}
+        canvas_graph_detail_ix = {}
+        for f, relpath in self._canvas_file_index.items():
+            content_c = get_canvas_content(
+                self._dirpath / relpath)
+            canvas_content_ix[f] = content_c
+            G_c, pos_c, edge_labels_c = get_canvas_graph_detail(
+                content_c)
+            canvas_graph_detail_ix[f] = G_c, pos_c, edge_labels_c
+        self._canvas_content_index = canvas_content_ix
+        self._canvas_graph_detail_index = canvas_graph_detail_ix
+
         return self  # fluent
 
-    def gather(self, *, tags=None):
+    def gather(self, *, tags: list[str] = None):
         """gather the content of your notes so that all the plaintext is
         stored in one place for easy access.
 
@@ -264,24 +419,22 @@ class Vault:
                 will remove all header formatting (e.g. '#', '##' chars)
                 and produces a one-line string.
         """
-        if not self._is_connected:
-            raise AttributeError('Connect vault before gathering notes.')
-
         # source text will not remove any content:
         self._source_text_index = {
-            k: _get_source_text_from_md_file(self._dirpath / v,
-                                             remove_code=False)
+            k: get_source_text_from_md_file(self._dirpath / v,
+                                            remove_code=True,
+                                            remove_math=True)
             for k, v in self._file_index.items()}
         self._readable_text_index = {
-            k: _get_readable_text_from_md_file(self._dirpath / v,
-                                               tags=tags)
+            k: get_readable_text_from_md_file(self._dirpath / v,
+                                              tags=tags)
             for k, v in self._file_index.items()}
 
         self._is_gathered = True
 
         return self  # fluent
 
-    def get_backlinks(self, note_name):
+    def get_backlinks(self, note_name: str) -> list[str]:
         """Get backlinks for a note (given its name).
 
         If a note has not been created, but has wikilinks pointing to it
@@ -305,7 +458,7 @@ class Vault:
         else:
             return self._backlinks_index[note_name]
 
-    def get_backlink_counts(self, note_name):
+    def get_backlink_counts(self, note_name: str) -> dict[str, int]:
         """Get counts of backlinks for a note (given its name).
 
         Args:
@@ -324,7 +477,7 @@ class Vault:
             backlinks = self.get_backlinks(note_name)
             return dict(Counter(backlinks))
 
-    def get_wikilinks(self, file_name):
+    def get_wikilinks(self, file_name: str) -> list[str]:
         """Get wikilinks for a note (given its filename).
 
         Wikilinks can only appear in notes that already exist, so if a
@@ -346,7 +499,26 @@ class Vault:
         else:
             return self._wikilinks_index[file_name]
 
-    def get_embedded_files(self, file_name):
+    def get_wikilink_counts(self, note_name: str) -> dict[str, int]:
+        """Get counts of wikilinks for a note (given its name).
+
+        Args:
+            note_name (str): the string that is the name in the graph.
+                This is NOT a filepath!
+
+        Returns:
+            dict of integers >= 1
+        """
+        if not self._is_connected:
+            raise AttributeError('Connect notes before calling the function')
+
+        if note_name not in self._graph.nodes:
+            raise ValueError('"{}" not found in graph.'.format(note_name))
+        else:
+            wikilinks = self.get_wikilinks(note_name)
+            return dict(Counter(wikilinks))
+
+    def get_embedded_files(self, file_name: str) -> list[str]:
         """Get embedded files for a note (given its filename).
 
         Embedded files can only appear in notes that already exist, so if a
@@ -368,7 +540,7 @@ class Vault:
         else:
             return self._embedded_files_index[file_name]
 
-    def get_md_links(self, file_name):
+    def get_md_links(self, file_name: str) -> list[str]:
         """Get markdown links for a note (given its filename).
 
         Markdown links can only appear in notes that already exist, so if a
@@ -390,7 +562,7 @@ class Vault:
         else:
             return self._md_links_index[file_name]
 
-    def get_front_matter(self, file_name):
+    def get_front_matter(self, file_name: str) -> list[dict]:
         """Get front matter for a note (given its filename).
 
         Front matter can only appear in notes that already exist, so if a
@@ -411,9 +583,11 @@ class Vault:
         else:
             return self._front_matter_index[file_name]
 
-    def get_tags(self, file_name):
+    def get_tags(self, file_name: str, *,
+                 show_nested: bool = False) -> list[str]:
         """Get tags for a note (given its filename).
-        Only top-level tags are supported (NOT nested tags).
+        By default, only the highest level of any nested tags is shown
+        in the output.
 
         Tags can only appear in notes that already exist, so if a
         note is not in the file_index at all then the function will raise
@@ -433,9 +607,9 @@ class Vault:
         else:
             return self._tags_index[file_name]
 
-    def get_source_text(self, file_name):
+    def get_source_text(self, file_name: str) -> str:
         """Get text for a note (given its filename).  This requires the vault
-        functions 'connect' AND 'gather' to have been called.
+        function 'gather' to have been called.
 
         Text can only appear in notes that already exist, so if a
         note is not in the file_index at all then the function will raise
@@ -455,10 +629,9 @@ class Vault:
         else:
             return self._source_text_index[file_name]
 
-    def get_readable_text(self, file_name):
+    def get_readable_text(self, file_name: str) -> str:
         """Get readable text for a note (given its filename).
-        This requires the vault functions 'connect' AND 'gather' to have
-        been called.
+        This requires the vault function 'gather' to have been called.
 
         Change the arguments of the 'gather' function to specify how the text
         output should be stored, e.g. whether all HTML tags should be removed
@@ -482,8 +655,8 @@ class Vault:
         else:
             return self._readable_text_index[file_name]
 
-    def _get_md_relpaths(self, **kwargs):
-        """Return list of filepaths *relative* to the directory instantiated
+    def _get_md_relpaths(self, **kwargs) -> list[Path]:
+        """Return list of md filepaths *relative* to the directory instantiated
         for the class.
 
         Returns:
@@ -491,80 +664,74 @@ class Vault:
         """
         return get_md_relpaths_matching_subdirs(self._dirpath, **kwargs)
 
-    def _get_md_relpaths_by_name(self, **kwargs):
+    def _get_canvas_relpaths(self, **kwargs) -> list[Path]:
+        """Return list of canvas filepaths *relative* to the directory instantiated
+        for the class.
+
+        Returns:
+            list
+        """
+        return get_canvas_relpaths_matching_subdirs(self._dirpath,
+                                                    **kwargs)
+
+    def __get_relpaths_by_name(self, *, extension, **kwargs) -> dict[str, Path]:
         """Return k,v pairs
         where k is the file name
-        and v is the relpath of the md file
+        and v is the relpath of the {extension} file
 
         Returns:
             dict
         """
-        return {f.stem: f for f in self._get_md_relpaths(**kwargs)}
+        # get note names (for k) and relpaths (for v):
+        # keep or remove ext, based on how the file refs appear in wikilinks:
+        if extension == 'md':
+            relpaths_list = self._get_md_relpaths(
+                **kwargs)
+            # remove .md ext:
+            all_file_names_list = [f.stem for f in relpaths_list]
+        if extension == 'canvas':
+            relpaths_list = self._get_canvas_relpaths(
+                **kwargs)
+            # keep .canvas ext:
+            all_file_names_list = [f.name for f in relpaths_list]
 
-    def _get_wikilinks_index(self):
-        """Return k,v pairs
-        where k is the md filename
-        and v is list of ALL wikilinks found in k"""
-        return {k: get_wikilinks(self._dirpath / v)
-                for k, v in self._file_index.items()}
+        # get indices of dupe note names:
+        _, inverse_ix, counts = np.unique(
+            np.array(all_file_names_list),
+            return_inverse=True,
+            return_counts=True,
+            axis=0)
+        dupe_names_ix = np.where(counts[inverse_ix] > 1)[0]
 
-    def _get_embedded_files_index(self):
-        """Return k,v pairs
-        where k is the md filename
-        and v is list of ALL embedded files found in k"""
-        return {k: get_embedded_files(self._dirpath / v)
-                for k, v in self._file_index.items()}
+        # get shortest paths via mask:
+        shortest_paths_arr = np.array(all_file_names_list, dtype=object)
+        if extension == 'md':
+            shortest_paths_arr[dupe_names_ix] = np.array(
+                [str(fpath.with_suffix(''))
+                 for fpath in relpaths_list])[dupe_names_ix]
+        if extension == 'canvas':
+            shortest_paths_arr[dupe_names_ix] = np.array(
+                [str(fpath)
+                 for fpath in relpaths_list])[dupe_names_ix]
+        return {n: p for n, p in zip(shortest_paths_arr, relpaths_list)}
 
-    def _get_math_index(self):
-        """Return k,v pairs
-        where k is the md filename
-        and v is list of ALL LaTeX math strings found in k"""
-        return {k: _get_all_latex_from_md_file(self._dirpath / v)
-                for k, v in self._file_index.items()}
+    def _get_md_relpaths_by_name(self, **kwargs) -> dict[str, Path]:
+        return self.__get_relpaths_by_name(extension='md',
+                                           **kwargs)
 
-    def _get_unique_wikilinks_index(self):
-        """Return k,v pairs
-        where k is the md filename
-        and v is list of UNIQUE wikilinks found in k"""
-        return {k: get_unique_wikilinks(self._dirpath / v)
-                for k, v in self._file_index.items()}
+    def _get_canvas_relpaths_by_name(self, **kwargs) -> dict[str, Path]:
+        return self.__get_relpaths_by_name(extension='canvas',
+                                           **kwargs)
 
-    def _get_md_links_index(self):
-        """Return k,v pairs
-        where k is the md note name
-        and v is list of ALL markdown links found in k"""
-        return {k: get_md_links(self._dirpath / v)
-                for k, v in self._file_index.items()}
-
-    def _get_unique_md_links_index(self):
-        """Return k,v pairs
-        where k is the md note name
-        and v is list of UNIQUE markdown links found in k"""
-        return {k: get_unique_md_links(self._dirpath / v)
-                for k, v in self._file_index.items()}
-
-    def _get_backlinks_index(self, *, graph):
+    def _get_backlinks_index(self, *,
+                             graph: nx.MultiDiGraph) -> dict[str, list[str]]:
         """Return k,v pairs
         where k is the md note name
         and v is list of ALL backlinks found in k"""
         return {n: [n[0] for n in list(graph.in_edges(n))]
                 for n in self._graph.nodes}
 
-    def _get_front_matter_index(self):
-        """Return k,v pairs
-        where k is the md filename
-        and v is list of file matter metadata found in k"""
-        return {k: get_front_matter(self._dirpath / v)
-                for k, v in self._file_index.items()}
-
-    def _get_tags_index(self):
-        """Return k,v pairs
-        where k is the md filename
-        and v is list of tags found in k"""
-        return {k: get_tags(self._dirpath / v)
-                for k, v in self._file_index.items()}
-
-    def get_note_metadata(self):
+    def get_note_metadata(self) -> pd.DataFrame:
         """Structured dataset of metadata on the vault's notes.  This
         includes filepaths and counts of different link types.
 
@@ -587,7 +754,8 @@ class Vault:
               )
         return df
 
-    def _create_note_metadata_columns(self, df):
+    def _create_note_metadata_columns(self,
+                                      df: pd.DataFrame) -> pd.DataFrame:
         """pipe func for mutating df"""
         df['rel_filepath'] = [self._file_index.get(f, np.NaN)
                               for f in df.index]
@@ -602,6 +770,10 @@ class Vault:
                                      [len(self._wikilinks_index.get(f, []))
                                      for f in df.index],
                                      np.NaN)
+        df['n_tags'] = np.where(df['note_exists'],
+                                [len(self._tags_index.get(f, []))
+                                 for f in df.index],
+                                np.NaN)
         df['n_embedded_files'] = np.where(df['note_exists'],
                                           [len(self._embedded_files_index.get(
                                               f, []))
@@ -614,7 +786,8 @@ class Vault:
         )
         return df
 
-    def _clean_up_note_metadata_dtypes(self, df):
+    def _clean_up_note_metadata_dtypes(self,
+                                       df: pd.DataFrame) -> pd.DataFrame:
         """pipe func for mutating df"""
         df['rel_filepath'] = np.where(df['rel_filepath'].notna(),
                                       [Path(str(f))
@@ -623,7 +796,7 @@ class Vault:
         df['n_wikilinks'] = df['n_wikilinks'].astype(float)  # for consistency
         return df
 
-    def _get_nonexistent_notes(self):
+    def _get_nonexistent_notes(self) -> list[str]:
         """Get notes that have backlinks but don't have md files.
 
         The comparison is done with sets but the result is returned
@@ -631,7 +804,8 @@ class Vault:
         return list(set(self.backlinks_index.keys())
                     .difference(set(self.file_index)))
 
-    def _get_isolated_notes(self, *, graph):
+    def _get_isolated_notes(self, *,
+                            graph: nx.MultiDiGraph) -> list[str]:
         """Get notes that are not connected to any other notes in the vault,
         i.e. they have 0 wikilinks and 0 backlinks.
 
