@@ -4,20 +4,21 @@ from pathlib import Path
 from glob import glob
 import markdown
 import frontmatter
-from .html_processing import (_get_source_plaintext_from_html,
+from ._constants import (WIKILINK_REGEX,
+                         TAG_MAIN_ONLY_REGEX, TAG_INCLUDE_NESTED_REGEX,
+                         WIKILINK_AS_STRING_REGEX,
+                         EMBEDDED_FILE_LINK_AS_STRING_REGEX,
+                         INLINE_LINK_AFTER_HTML_PROC_REGEX,
+                         INLINE_LINK_VIA_MD_ONLY_REGEX)
+from ._io import (get_relpaths_from_dir,
+                  get_relpaths_matching_subdirs)
+from .html_processing import (_get_plaintext_from_html,
                               _remove_code, _remove_latex, _remove_del_text,
                               _remove_main_formatting,
                               _get_all_latex_from_html_content)
 
-# wikilink & embedded file regex: regex that includes any aliases
-# group 0 captures embedded link; group 1 is everything inside [[]]
-WIKILINK_REGEX = r'(!)?\[{2}([^\]\]]+)\]{2}'
-# md links regex: catch URLs or paths
-INLINE_LINK_AFTER_HTML_PROC_REGEX = r'\[([^\]]+)\]\(<([^)]+)>\)'
-INLINE_LINK_VIA_MD_ONLY_REGEX = r'\[([^\]]+)\]\(([^)]+)\)'
 
-
-def get_md_relpaths_from_dir(dir_path):
+def get_md_relpaths_from_dir(dir_path: Path) -> list[Path]:
     """Get list of relative paths for markdown files in a given directory,
     including any subdirectories.
 
@@ -31,12 +32,12 @@ def get_md_relpaths_from_dir(dir_path):
     Returns:
         list of Path objects
     """
-    return [Path(p).relative_to(dir_path)
-            for p in glob(f"{dir_path}/**/*.md", recursive=True)]
+    return get_relpaths_from_dir(dir_path, extension='md')
 
 
-def get_md_relpaths_matching_subdirs(dir_path, *,
-                                     include_subdirs=None, include_root=True):
+def get_md_relpaths_matching_subdirs(dir_path: Path, *,
+                                     include_subdirs: list = None,
+                                     include_root: bool = True) -> list[Path]:
     """Get list of relative paths for markdown files in a given directory,
     filtered to include specified subdirectories (with include_subdirs
     kwarg).  The default arguments align with get_md_relpaths_from_dir
@@ -64,27 +65,14 @@ def get_md_relpaths_matching_subdirs(dir_path, *,
     Returns:
         list of Path objects
     """
-    if include_subdirs:
-        include_subdirs_final = [str(Path(i).as_posix())
-                                 for i in include_subdirs]
-
-    if not include_subdirs and include_root:
-        return get_md_relpaths_from_dir(dir_path)
-    elif not include_subdirs and not include_root:
-        return [i for i in get_md_relpaths_from_dir(dir_path)
-                if str(i.parent.as_posix()) != '.']
-    else:
-        if include_root:
-            return [i for i in get_md_relpaths_from_dir(dir_path)
-                    if str(i.parent.as_posix())
-                    in include_subdirs_final + ['.']]
-        else:
-            return [i for i in get_md_relpaths_from_dir(dir_path)
-                    if str(i.parent.as_posix())
-                    in include_subdirs_final]
+    return get_relpaths_matching_subdirs(
+        dir_path,
+        extension='md',
+        include_subdirs=include_subdirs,
+        include_root=include_root)
 
 
-def get_wikilinks(filepath):
+def get_wikilinks(filepath: Path) -> list[str]:
     """Get ALL wikilinks from a md file.
     The links' order of appearance in the file IS preserved in the output.
 
@@ -103,14 +91,14 @@ def get_wikilinks(filepath):
     Returns:
         list of strings
     """
-    src_txt = _get_source_text_from_md_file(filepath, remove_code=True)
+    src_txt = get_source_text_from_md_file(filepath, remove_code=True)
 
     wikilinks = _get_all_wikilinks_from_source_text(
         src_txt, remove_aliases=True)
     return wikilinks
 
 
-def get_embedded_files(filepath):
+def get_embedded_files(filepath: Path) -> list[str]:
     """Get ALL embedded files from a md file.
     The embedded files' order of appearance in the file IS preserved in the output.
 
@@ -127,14 +115,14 @@ def get_embedded_files(filepath):
     Returns:
         list of strings
     """
-    src_txt = _get_source_text_from_md_file(filepath, remove_code=True)
+    src_txt = get_source_text_from_md_file(filepath, remove_code=True)
 
     files = _get_all_embedded_files_from_source_text(
         src_txt, remove_aliases=True)
     return files
 
 
-def get_unique_wikilinks(filepath):
+def get_unique_wikilinks(filepath: Path) -> list[str]:
     """Get UNIQUE wikilinks from a md file.
     The links' order of appearance in the file IS preserved in the output.
 
@@ -151,13 +139,13 @@ def get_unique_wikilinks(filepath):
     Returns:
         list of strings
     """
-    src_txt = _get_source_text_from_md_file(filepath, remove_code=True)
+    src_txt = get_source_text_from_md_file(filepath, remove_code=True)
 
-    wikilinks = _get_unique_wikilinks(src_txt, remove_aliases=True)
+    wikilinks = _get_unique_wikilinks_from_source_text(src_txt, remove_aliases=True)
     return wikilinks
 
 
-def get_md_links(filepath):
+def get_md_links(filepath: Path) -> list[str]:
     """Get markdown links from a md file.
     The links' order of appearance in the file IS preserved in the output.
 
@@ -172,8 +160,11 @@ def get_md_links(filepath):
     Returns:
         list of strings
     """
-    src_txt = _get_source_text_from_md_file(filepath, remove_code=True)
+    src_txt = get_source_text_from_md_file(filepath, remove_code=True)
+    return _get_md_links_from_source_text(src_txt)
 
+
+def _get_md_links_from_source_text(src_txt: str) -> list[str]:
     links = _get_all_md_link_info_from_source_text(src_txt)
     if links:  # return links only, not their text
         return [t[-1] for t in links]
@@ -181,7 +172,7 @@ def get_md_links(filepath):
         return links  # empty list
 
 
-def get_unique_md_links(filepath):
+def get_unique_md_links(filepath: Path) -> list[str]:
     """Get markdown links (unique) from a md file.
     The links' order of appearance in the file IS preserved in the output.
 
@@ -196,13 +187,13 @@ def get_unique_md_links(filepath):
     Returns:
         list of strings
     """
-    src_txt = _get_source_text_from_md_file(filepath, remove_code=True)
+    src_txt = get_source_text_from_md_file(filepath, remove_code=True)
 
     links = _get_unique_md_links_from_source_text(src_txt)
     return links
 
 
-def get_front_matter(filepath):
+def get_front_matter(filepath: Path) -> dict:
     """Get front matter from a md file.
 
     If no front matter is found for a file, the value will be {}.
@@ -217,29 +208,34 @@ def get_front_matter(filepath):
     return front_matter
 
 
-def get_tags(filepath):
+def get_tags(filepath: Path, *, show_nested: bool = False) -> list[str]:
     """Get tags from a md file, based on the order they appear in the file.
-    Only top-level tags are extracted: nested tag detail is NOT supported.
+    By default, only the highest level of any nested tags is shown in the
+    output.
 
     If no tags are found for a file, the value will be [].
 
     Args:
         filepath (pathlib Path): Path object representing the file from
             which info will be extracted.
+        show_nested (Boolean): show nested tags in the output.  Defaults to
+            False (which would mean only the highest level of any nested tags
+            are included in the output).
     Returns:
         list
     """
     # get text from source file, but remove any '\#' and code:
-    src_txt = _get_source_text_from_md_file(
+    src_txt = get_source_text_from_md_file(
         filepath, remove_code=True,
         str_transform_func=_transform_md_file_string_for_tag_parsing)
     # remove wikilinks so that '#' headers are not caught:
     src_txt = _remove_wikilinks_from_source_text(src_txt)
-    tags = _get_tags_from_source_text(src_txt)
+    tags = _get_tags_from_source_text(src_txt, show_nested=show_nested)
     return tags
 
 
-def _get_md_front_matter_and_content(filepath, *, str_transform_func=None):
+def _get_md_front_matter_and_content(filepath: Path, *,
+                                     str_transform_func=None) -> tuple[dict, str]:
     """parse md file into front matter and note content"""
     with open(filepath, encoding='utf-8') as f:
         try:
@@ -248,7 +244,11 @@ def _get_md_front_matter_and_content(filepath, *, str_transform_func=None):
                 file_string = str_transform_func(file_string)
             return frontmatter.parse(file_string)
         # for invalid YAML, return the whole file as content:
-        except yaml.scanner.ScannerError:
+        except yaml.scanner.ScannerError as e:
+            print(f"Front matter not populated for {filepath.name}: {repr(e)}")
+            return {}, file_string
+        except yaml.parser.ParserError as e:
+            print(f"Front matter not populated for {filepath.name}: {repr(e)}")
             return {}, file_string
         # handle template {{}} chars in front matter:
         except yaml.constructor.ConstructorError:
@@ -256,20 +256,31 @@ def _get_md_front_matter_and_content(filepath, *, str_transform_func=None):
                 str.maketrans({"{": r"\{",
                                "}": r"\}"}))
             return frontmatter.parse(file_string_esc)
+        # any others:
+        except:
+            return {}, file_string
 
 
-def _get_html_from_md_file(filepath, *, str_transform_func=None):
+def _get_html_from_md_file(filepath: Path, *,
+                           str_transform_func=None) -> str:
     """md file -> html (without front matter).
 
     pymarkdown extensions are used and configured to reflect the Obsidian
     experience as much as possible.  For example, arithmatex is necessary
     to parse math, md_mermaid for mermaid diagram support, etc.
     """
-    _, content = _get_md_front_matter_and_content(
+    _, md_content = _get_md_front_matter_and_content(
         filepath,
         str_transform_func=str_transform_func)
-    html = markdown.markdown(content, output_format='html',
+    html = _get_html_from_md_content(md_content)
+    return html
+
+
+def _get_html_from_md_content(md_content: str) -> str:
+    """md content -> html (without front matter)"""
+    html = markdown.markdown(md_content, output_format='html',
                              extensions=['pymdownx.arithmatex',
+                                         'pymdownx.superfences',
                                          'pymdownx.mark',
                                          'pymdownx.tilde',
                                          'pymdownx.saneheaders',
@@ -281,19 +292,33 @@ def _get_html_from_md_file(filepath, *, str_transform_func=None):
     return html
 
 
-def _get_source_text_from_md_file(filepath, *,
-                                  remove_code=False, str_transform_func=None):
+def get_source_text_from_html(html: str, *,
+                              remove_code: bool = False,
+                              remove_math: bool = False) -> str:
+    """html (without front matter) -> ASCII plaintext"""
+    if remove_code:
+        html = _remove_code(html)
+    if remove_math:
+        html = _remove_latex(html)
+    return _get_plaintext_from_html(html)
+
+
+def get_source_text_from_md_file(filepath: Path, *,
+                                 remove_code: bool = False,
+                                 remove_math: bool = False,
+                                 str_transform_func=None) -> str:
     """md file -> html (without front matter) -> ASCII plaintext"""
     # strip out front matter (if any):
     html = _get_html_from_md_file(
         filepath,
         str_transform_func=str_transform_func)
-    if remove_code:
-        html = _remove_code(html)
-    return _get_source_plaintext_from_html(html)
+
+    return get_source_text_from_html(html, remove_code=remove_code,
+                                     remove_math=remove_math)
 
 
-def _get_readable_text_from_md_file(filepath, *, tags=None):
+def get_readable_text_from_md_file(filepath: Path, *,
+                                   tags: list[str] = None) -> str:
     """md file -> html -> plaintext with major formatting removed."""
     # strip out front matter (if any):
     html = _get_html_from_md_file(
@@ -311,10 +336,10 @@ def _get_readable_text_from_md_file(filepath, *, tags=None):
     else:  # defaults
         html = _remove_main_formatting(html)
 
-    return _get_source_plaintext_from_html(html)
+    return _get_plaintext_from_html(html)
 
 
-def _get_all_wikilinks_and_embedded_files(src_txt):
+def _get_all_wikilinks_and_embedded_files(src_txt: str) -> list[str]:
     # extract links
     pattern = re.compile(WIKILINK_REGEX)
 
@@ -322,14 +347,15 @@ def _get_all_wikilinks_and_embedded_files(src_txt):
     return link_matches_list
 
 
-def _remove_aliases_from_wikilink_regex_matches(link_matches_list):
+def _remove_aliases_from_wikilink_regex_matches(link_matches_list: list[str]) -> list[str]:
     return [(i.replace('\\', '')
              .split("|")[0].rstrip()  # catch alias/alt-text
              .split('#', 1)[0])  # catch links to headers
             for i in link_matches_list]
 
 
-def _get_all_wikilinks_from_source_text(src_txt, *, remove_aliases=True):
+def _get_all_wikilinks_from_source_text(src_txt: str, *,
+                                        remove_aliases: bool = True) -> list[str]:
     matches_list = _get_all_wikilinks_and_embedded_files(src_txt)
     link_matches_list = [g[1] for g in matches_list
                          if g[0] == '']
@@ -337,11 +363,15 @@ def _get_all_wikilinks_from_source_text(src_txt, *, remove_aliases=True):
     if remove_aliases:
         link_matches_list = _remove_aliases_from_wikilink_regex_matches(
             link_matches_list)
+
+    # remove .md:
+    link_matches_list = [name.removesuffix('.md')
+                         for name in link_matches_list]
     return link_matches_list
 
 
-def _get_all_embedded_files_from_source_text(src_txt, *,
-                                             remove_aliases=True):
+def _get_all_embedded_files_from_source_text(src_txt: str, *,
+                                             remove_aliases: bool = True) -> list[str]:
     matches_list = _get_all_wikilinks_and_embedded_files(src_txt)
     embedded_files_sublist = [g[1] for g in matches_list
                               if g[0] == '!']
@@ -352,47 +382,51 @@ def _get_all_embedded_files_from_source_text(src_txt, *,
     return embedded_files_sublist
 
 
-def _get_all_latex_from_md_file(filepath):
+def _get_all_latex_from_md_file(filepath: Path) -> list[str]:
     return _get_all_latex_from_html_content(
         _get_html_from_md_file(filepath))
 
 
-def _get_unique_wikilinks(src_txt, *, remove_aliases=True):
+def _get_unique_wikilinks_from_source_text(src_txt: str, *,
+                                           remove_aliases: bool = True) -> list[str]:
     wikilinks = _get_all_wikilinks_from_source_text(
         src_txt, remove_aliases=remove_aliases)
     return list(dict.fromkeys(wikilinks))
 
 
-def _get_all_md_link_info_from_source_text(src_txt):
+def _get_all_md_link_info_from_source_text(src_txt: str) -> list[tuple[str]]:
     links_regex = re.compile(INLINE_LINK_AFTER_HTML_PROC_REGEX)
 
     links_list_of_tuples = list(links_regex.findall(src_txt))
     return links_list_of_tuples
 
 
-def _get_unique_md_links_from_source_text(src_txt):
+def _get_unique_md_links_from_source_text(src_txt: str) -> list[str]:
     links_detail = _get_all_md_link_info_from_source_text(
         src_txt)
     links_list = [link for _, link in links_detail]
     return list(dict.fromkeys(links_list))
 
 
-def _remove_wikilinks_from_source_text(src_txt):
+def _remove_wikilinks_from_source_text(src_txt: str) -> str:
     return re.sub(WIKILINK_REGEX, '', src_txt)
 
 
-def _transform_md_file_string_for_tag_parsing(txt):
+def _transform_md_file_string_for_tag_parsing(txt: str) -> str:
     return txt.replace('\\#', '')
 
 
-def _get_tags_from_source_text(src_txt):
-    tags_regex = r'(?<!\()#{1}([A-z]+[0-9_\-]*[A-Z0-9]?)\/?'
-    pattern = re.compile(tags_regex)
+def _get_tags_from_source_text(src_txt: str, *,
+                               show_nested: bool = False) -> list[str]:
+    if not show_nested:
+        pattern = re.compile(TAG_MAIN_ONLY_REGEX)
+    else:
+        pattern = re.compile(TAG_INCLUDE_NESTED_REGEX)
     tags_list = pattern.findall(src_txt)
     return tags_list
 
 
-def _replace_wikilinks_with_their_text(src_txt):
+def _replace_wikilinks_with_their_text(src_txt: str) -> str:
     # get list of wikilinks as strings:
     links_list = _get_all_wikilinks_from_source_text(
         src_txt, remove_aliases=False)
@@ -414,9 +448,9 @@ def _replace_wikilinks_with_their_text(src_txt):
     return out_str
 
 
-def _replace_md_links_with_their_text(src_txt):
+def _replace_md_links_with_their_text(src_txt: str) -> str:
     # get list of wikilinks as strings:
-    matched_text_list = re.findall(r'\[[^\]]+\]\([^)]+\)', src_txt)
+    matched_text_list = re.findall(WIKILINK_AS_STRING_REGEX, src_txt)
     # get the detail from groups:
     links_detail = re.findall(INLINE_LINK_VIA_MD_ONLY_REGEX, src_txt)
 
@@ -432,9 +466,9 @@ def _replace_md_links_with_their_text(src_txt):
     return out_str
 
 
-def _remove_embedded_file_links_from_text(src_txt):
+def _remove_embedded_file_links_from_text(src_txt: str) -> str:
     # get list of embedded file links as strings:
-    links_list = re.findall(r'!?\[{2}([^\]\]]+)\]{2}', src_txt)
+    links_list = re.findall(EMBEDDED_FILE_LINK_AS_STRING_REGEX, src_txt)
     # add in the ![[...]] chars:
     links_list = ["".join(['![[', i, ']]']) for i in links_list]
 
